/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.konfiguration;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.ReferenceArray;
import de.bsvrz.dav.daf.main.Data.ReferenceValue;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.NonMutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * <code>Konfiguration</code> verwaltet Konfigurationsdaten - Straßensegmente
 * und -Knoten, Messquerschnitte und Straßenabschnitte. Die Klasse ist genau ein
 * Konfigurationsobjekt vom Typ typ.netz und für eine DAV-Verbindung
 * instanziierbar. Bei der Instanziierung werden die Straßensegmente und
 * -Knoten, Messquerschnitte und Straßenabschnitte für das übergebene Netz
 * geladen und für diese Objekttypen wird das DAV-Datenmodell abgebildet.
 * 
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class Konfiguration {
	/**
	 * Debug-Logger für Logging-Ausgaben.
	 */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Erzeugt eine Instanz vom Typ <code>Konfiguration</code> für eine
	 * DAV-Verbindung und ein Konfigurationsobjekt vom Typ typ.netz.
	 * 
	 * @param connection
	 *            die Datenverteiler-Verbindung
	 * @param netz
	 *            das (Konfigurations-)Objekt, dessen Straßensegmente und
	 *            -Knoten, Messquerschnitte und Straßenabschnitte ausgewertet
	 *            werden
	 * @param nurKorrekteTmcInfo
	 *            wenn <code>true</code> werden äußere Strassensegmente mit
	 *            fehlerhafter oder unvollstängen TMC-Information nicht
	 *            übernommen
	 * @param keineInnerenSegmenteAussen
	 *            wenn <code>true</code> werden innere Strassensegmente, die
	 *            "aussen" liegen (d. h. deren Vorgänger/Nachfolger nicht mehr
	 *            im Netz liegt) nicht übernommen
	 */
	public Konfiguration(final ClientDavInterface connection, final ConfigurationObject netz,
			final boolean nurKorrekteTmcInfo, final boolean keineInnerenSegmenteAussen) {
		_connection = connection;
		_netz = netz;

		_strassenKnoten = new HashMap<String, StrassenKnoten>();
		_tmcPunkte = new HashMap<String, TmcLocationPunkt>();
		_tmcPunkteMitLocationCode = new HashMap<String, TmcLocationPunkt>();

		aeussereStrassenSegmente = bestimmeAeussereStrassenSegmente(connection, netz, nurKorrekteTmcInfo);
		_debug.config(netz.getPid() + ": " + aeussereStrassenSegmente.size() + " äußere(s) Straßensegment(e) mit "
				+ _tmcPunkte.size() + " TMC-Punkt(en) und " + _strassenKnoten.size()
				+ " Straßenknoten werden bearbeitet");

		_innereStrassenSegmente = bestimmeInnereStrassenSegmente(connection, aeussereStrassenSegmente,
				keineInnerenSegmenteAussen);
		_debug.config(netz.getPid() + ": " + _innereStrassenSegmente.size()
				+ " innere(s) Straßensegment(e) werden bearbeitet");

		_strassenTeilSegmente = bestimmeStrassenTeilSegmente(connection, aeussereStrassenSegmente);
		_strassenTeilSegmente.putAll(bestimmeStrassenTeilSegmente(connection, _innereStrassenSegmente));
		_debug.config(netz.getPid() + ": " + _strassenTeilSegmente.size() + " Straßenteilsegment(e) werden bearbeitet");

		_messQuerschnitte = bestimmeMessQuerschnitte(connection, aeussereStrassenSegmente);
		_messQuerschnitte.putAll(bestimmeMessQuerschnitte(connection, _innereStrassenSegmente));
		_debug.config(netz.getPid() + ": " + _messQuerschnitte.size() + " Messquerschnitt(e) werden bearbeitet");

		_strassenAbschnitte = bestimmeStrassenAbschnitte(connection, _messQuerschnitte);
		_debug.config(netz.getPid() + ": " + _strassenAbschnitte.size() + " Straßenabschnitt(e) werden bearbeitet");
	}

	/**
	 * Erzeugt eine Instanz vom Typ <code>Konfiguration</code> für eine
	 * DAV-Verbindung und ein Konfigurationsobjekt vom Typ typ.netz.
	 * 
	 * @param connection
	 *            die Datenveteiler-Verbindung
	 * @param netz
	 *            das (Konfigurations-)Objekt, dessen Straßensegmente und
	 *            -Knoten, Messquerschnitte und Straßenabschnitte ausgewertet
	 *            werden
	 */
	public Konfiguration(final ClientDavInterface connection, final ConfigurationObject netz) {
		this(connection, netz, false, false);
	}

	/**
	 * DAV-Verbindung.
	 */
	private ClientDavInterface _connection;

	/**
	 * 
	 * @return die DAV-Verbindung.
	 */
	private ClientDavInterface getConnection() {
		return _connection;
	}

	/**
	 * Konfigurationsobjekt vom Typ typ.netz.
	 */
	private ConfigurationObject _netz;

	/**
	 * 
	 * @return das Konfigurationsobjekt vom Typ typ.netz.
	 */
	public ConfigurationObject getNetz() {
		return _netz;
	}

	/**
	 * Die Straßenknoten.
	 */
	private Map<String, StrassenKnoten> _strassenKnoten;

	/**
	 * Read-only Zugriff auf die Straßenknoten.
	 * 
	 * @return die Straßenknoten
	 */
	public Map<String, StrassenKnoten> getStrassenKnoten() {
		return Collections.unmodifiableMap(_strassenKnoten);
	}

	/**
	 * Die Tmc-Punkte.
	 */
	private Map<String, TmcLocationPunkt> _tmcPunkte;

	/**
	 * Read-only Zugriff auf die Tmc-Punkte.
	 * 
	 * @return Zuordnung der Tmc-Punkte zu ihrer PID
	 */
	public Map<String, TmcLocationPunkt> getTmcPunkte() {
		return Collections.unmodifiableMap(_tmcPunkte);
	}

	/**
	 * Die Tmc-Punkte, Zugriff über Locationcode.
	 */
	private Map<String, TmcLocationPunkt> _tmcPunkteMitLocationCode;

	/**
	 * Read-only Zugriff auf die Tmc-Punkte mit Zugriff über Locationcode.
	 * 
	 * @return Zuordnung der Tmc-Punkte zu ihrem Locationcode
	 */
	public Map<String, TmcLocationPunkt> getTmcPunkteMitLocationCode() {
		return Collections.unmodifiableMap(_tmcPunkteMitLocationCode);
	}

	/**
	 * Die äußeren Straßensegmente.
	 */
	private Map<String, AeusseresStrassenSegment> aeussereStrassenSegmente;

	/**
	 * Read-only Zugriff auf die äußeren Straßensegmente.
	 * 
	 * @return die äußeren Straßensegmente
	 */
	public Map<String, AeusseresStrassenSegment> getAeussereStrassenSegmente() {
		return Collections.unmodifiableMap(aeussereStrassenSegmente);
	}

	/**
	 * Die inneren Straßensegmente.
	 */
	private Map<String, InneresStrassenSegment> _innereStrassenSegmente;

	/**
	 * Read-only Zugriff auf die inneren Straßensegmente.
	 * 
	 * @return die inneren Straßensegmente
	 */
	public Map<String, InneresStrassenSegment> getInnereStrassenSegmente() {
		return Collections.unmodifiableMap(_innereStrassenSegmente);
	}

	/**
	 * Die relevanten Messquerschnitte.
	 */
	private Map<String, MessQuerschnitt> _messQuerschnitte;

	/**
	 * Read-only Zugriff auf die Messquerschnitte.
	 * 
	 * @return die relevanten Messquerschnitte
	 */
	public Map<String, MessQuerschnitt> getMessQuerschnitte() {
		return Collections.unmodifiableMap(_messQuerschnitte);
	}

	/**
	 * Die relevanten Strassenteilsegmente.
	 */
	private Map<String, StrassenTeilSegment> _strassenTeilSegmente;

	/**
	 * Read-only Zugriff auf die Strassenteilsegmente.
	 * 
	 * @return die relevanten Strassenteilsegmente
	 */
	public Map<String, StrassenTeilSegment> getStrassenTeilSegmente() {
		return Collections.unmodifiableMap(_strassenTeilSegmente);
	}

	/**
	 * Die relevanten Straßenabschnitte.
	 */
	private Map<String, StrassenAbschnitt> _strassenAbschnitte;

	/**
	 * Read-only Zugriff auf die Messquerschnitte.
	 * 
	 * @return die relevanten Straßenabschnitte
	 */
	public Map<String, StrassenAbschnitt> getStrassenAbschnitte() {
		return Collections.unmodifiableMap(_strassenAbschnitte);
	}

	/**
	 * Bestimmt alle äusseren Strassensegmente im übergebenen Netz und dessen
	 * Unternetzen.
	 * 
	 * @param connection
	 *            {@link ClientDavInterface}
	 * @param netz
	 *            das (Wurzel-) Netz, in dem die äusseren Strassensegmente
	 *            bestimmt werden
	 * @param nurKorrekteTmcInfo
	 *            nmur korrekte TMC-Infos
	 * @return Liste alle gefundenen äusseren Strassensegmente
	 */
	private Map<String, AeusseresStrassenSegment> bestimmeAeussereStrassenSegmente(final ClientDavInterface connection,
			final ConfigurationObject netz, final boolean nurKorrekteTmcInfo) {
		if (!netz.isOfType("typ.netz")) {
			throw new IllegalArgumentException("Objekt " + netz + " ist nicht vom Typ typ.netz");
		}

		final NonMutableSet menge = netz.getNonMutableSet("NetzBestandTeile");
		if (menge != null) {
			final Map<String, AeusseresStrassenSegment> result = new HashMap<String, AeusseresStrassenSegment>();

			for (SystemObject obj : menge.getElements()) {
				if (obj.isOfType("typ.äußeresStraßenSegment")) {
					AttributeGroup atg = connection.getDataModel().getAttributeGroup("atg.straßenSegment");

					Data data = obj.getConfigurationData(atg);
					if (data != null) {
						final long laenge = data.getItem("Länge").asUnscaledValue().longValue() / 100;

						final SystemObject strasse = data.getItem("gehörtZuStraße").asReferenceValue()
								.getSystemObject();
						String nummer = null;
						if (strasse != null) {
							atg = connection.getDataModel().getAttributeGroup("atg.straße");
							data = strasse.getConfigurationData(atg);
							if (data != null) {
								nummer = data.getItem("Nummer").asTextValue().getText();
							}
							_debug.finest("Straßennummer von " + obj.getPid() + ": " + nummer);
						}
						atg = connection.getDataModel().getAttributeGroup("atg.äußeresStraßenSegment");
						data = obj.getConfigurationData(atg);
						if (data != null) {
							final SystemObject von = data.getItem("vonKnoten").asReferenceValue().getSystemObject();
							final SystemObject nach = data.getItem("nachKnoten").asReferenceValue().getSystemObject();

							final int tmcRichtung = data.getItem("TmcRichtung").asUnscaledValue().intValue();
							final SystemObject tmcPrimaerOrt = data.getItem("TmcPrimärOrt").asReferenceValue()
									.getSystemObject();
							TmcLocationPunkt tmcPunkt = null;

							if (tmcPrimaerOrt != null) {
								tmcPunkt = _tmcPunkte.get(tmcPrimaerOrt.getPid());
								if (tmcPunkt == null) {
									tmcPunkt = new TmcLocationPunkt(tmcPrimaerOrt, connection.getDataModel());
									_tmcPunkte.put(tmcPrimaerOrt.getPid(), tmcPunkt);
									_tmcPunkteMitLocationCode.put("" + tmcPunkt.getLocationCode(), tmcPunkt);
								}
								_debug.finest(
										"TMC Location Code von " + obj.getPid() + ": " + tmcPunkt.getLocationCode());

								if (tmcPunkt.getVorgaenger() == null || tmcPunkt.getNachfolger() == null) {
									atg = connection.getDataModel().getAttributeGroup("atg.tmcPunkt");
									data = tmcPrimaerOrt.getConfigurationData(atg);

									if (tmcPunkt.getVorgaenger() == null) {
										final SystemObject vorg = data.getItem("Vorgänger").asReferenceValue()
												.getSystemObject();
										if (vorg != null) {
											TmcLocationPunkt vorgaenger = _tmcPunkte.get(vorg.getPid());
											if (vorgaenger == null) {
												vorgaenger = new TmcLocationPunkt(vorg, connection.getDataModel());
												_tmcPunkte.put(vorg.getPid(), vorgaenger);
												_tmcPunkteMitLocationCode.put("" + vorgaenger.getLocationCode(),
														vorgaenger);
											}
											tmcPunkt.setVorgaenger(vorgaenger);
											if (vorgaenger.getNachfolger() == null) {
												vorgaenger.setNachfolger(tmcPunkt);
											} else if (vorgaenger.getNachfolger() != tmcPunkt) {
												_debug.warning("Inkonsistente TMC-Location Daten: TMC-Knoten "
														+ vorgaenger.getLocationCode() + " wurde als Vorgänger von "
														+ tmcPunkt.getLocationCode() + " bestimmt, hat aber bereits "
														+ vorgaenger.getNachfolger().getLocationCode()
														+ " als Nachfolger");
											}
										}
									}
									if (tmcPunkt.getNachfolger() == null) {
										final SystemObject nachf = data.getItem("Nachfolger").asReferenceValue()
												.getSystemObject();
										if (nachf != null) {
											TmcLocationPunkt nachfolger = _tmcPunkte.get(nachf.getPid());
											if (nachfolger == null) {
												nachfolger = new TmcLocationPunkt(nachf, connection.getDataModel());
												_tmcPunkte.put(nachf.getPid(), nachfolger);
												_tmcPunkteMitLocationCode.put("" + nachfolger.getLocationCode(),
														nachfolger);
											}
											tmcPunkt.setNachfolger(nachfolger);
											if (nachfolger.getVorgaenger() == null) {
												nachfolger.setVorgaenger(tmcPunkt);
											} else if (nachfolger.getVorgaenger() != tmcPunkt) {
												_debug.warning("Inkonsistente TMC-Location Daten: TMC-Knoten "
														+ nachfolger.getLocationCode() + " wurde als Nachfolger von "
														+ tmcPunkt.getLocationCode() + " bestimmt, hat aber bereits "
														+ nachfolger.getVorgaenger().getLocationCode()
														+ " als Vorgaenger");
											}
										}
									}
								}
							}

							if (!nurKorrekteTmcInfo || (tmcPunkt != null && tmcRichtung != 0)) {
								final AeusseresStrassenSegment segment = new AeusseresStrassenSegment(obj, laenge,
										tmcPunkt, tmcRichtung, nummer);

								if (von != null) {
									StrassenKnoten knoten = _strassenKnoten.get(von.getPid());
									if (knoten == null) {
										atg = connection.getDataModel().getAttributeGroup("atg.straßenKnoten");
										data = von.getConfigurationData(atg);
										if (data != null) {
											final String typ = data.getItem("Typ").asTextValue().getText();
											knoten = new StrassenKnoten(von, typ);
											_strassenKnoten.put(von.getPid(), knoten);
										}
									}
									segment.setVonKnoten(knoten);
									if (knoten != null) {
										knoten.addAbgehendesSegment(segment);
									}
								}

								if (nach != null) {
									StrassenKnoten knoten = _strassenKnoten.get(nach.getPid());
									if (knoten == null) {
										atg = connection.getDataModel().getAttributeGroup("atg.straßenKnoten");
										data = nach.getConfigurationData(atg);
										if (data != null) {
											final String typ = data.getItem("Typ").asTextValue().getText();
											knoten = new StrassenKnoten(nach, typ);
											_strassenKnoten.put(nach.getPid(), knoten);
										}
									}
									segment.setNachKnoten(knoten);
									if (knoten != null) {
										knoten.addHinfuehrendesSegment(segment);
									}
								}

								if (true /*
											 * obj.getPid().equals(
											 * "ass.p.11905.p.11906.0") ||
											 * obj.getPid().equals(
											 * "ass.n.11983.n.11982.0")
											 */) {
									result.put(obj.getPid(), segment);
								}
							} else {
								_debug.warning("Korrekte TMC-Daten (Ort und/oder Richtung ) nicht versorgt für " + obj
										+ "\n=> KANN NICHT VERWENDET WERDEN");
							}
						} else {
							_debug.warning("Daten von " + atg + " nicht versorgt für " + obj
									+ "\n=> KANN NICHT VERWENDET WERDEN");
						}
					} else {
						_debug.warning(
								"Daten von " + atg + " nicht versorgt für " + obj + "\n=> KANN NICHT VERWENDET WERDEN");
					}
				} else if (obj.isOfType("typ.netz")) {
					result.putAll(bestimmeAeussereStrassenSegmente(connection, (ConfigurationObject) obj,
							nurKorrekteTmcInfo));
				} else {
					_debug.warning("Objekt " + obj + " mit unerwartetem Typ " + obj.getType()
							+ " gefunden in der Menge \"NetzBestandTeile\" von " + netz);
				}
			}

			_debug.fine(result.size() + " äussere(s) Straßensegment(e) bestimmt in (Teil-) Netz " + netz.getPidOrId());
			return result;
		} else {
			throw new RuntimeException("Menge \"NetzBestandteile\" nicht vorhanden in " + netz);
		}
	}

	/**
	 * Bestimmt alle inneren Straßensegmente auf einer Menge von äußeren
	 * Straßensegmenten.
	 * 
	 * @param connection
	 *            DAV-Verbindung
	 * @param ass
	 *            äussere Straßensegmente, zu dennen die inneren Straßensegmente
	 *            bestimmt werden sollen
	 * @param keineInnerenSegmenteAussen
	 *            ???
	 * @return Sammlung aller gefundenen innerern Straßensegmente
	 */
	private Map<String, InneresStrassenSegment> bestimmeInnereStrassenSegmente(final ClientDavInterface connection,
			final Map<String, AeusseresStrassenSegment> ass, final boolean keineInnerenSegmenteAussen) {
		final Map<String, InneresStrassenSegment> result = new HashMap<String, InneresStrassenSegment>();
		final SystemObjectType type = connection.getDataModel().getType("typ.inneresStraßenSegment");

		final AttributeGroup atgSS = connection.getDataModel().getAttributeGroup("atg.straßenSegment");
		final AttributeGroup atgISS = connection.getDataModel().getAttributeGroup("atg.inneresStraßenSegment");

		connection.getDataModel().getConfigurationData(type.getElements(), atgSS);
		connection.getDataModel().getConfigurationData(type.getElements(), atgISS);

		for (SystemObject obj : type.getElements()) {
			if (obj.isValid()) {
				Data data = obj.getConfigurationData(atgSS);

				if (data != null) {
					final long laenge = data.getItem("Länge").asUnscaledValue().longValue() / 100;

					data = obj.getConfigurationData(atgISS);
					if (data != null) {
						final SystemObject von = data.getItem("vonStraßenSegment").asReferenceValue().getSystemObject();
						final SystemObject nach = data.getItem("nachStraßenSegment").asReferenceValue()
								.getSystemObject();
						final AeusseresStrassenSegment vonSegment = von != null ? ass.get(von.getPid()) : null;
						final AeusseresStrassenSegment nachSegment = nach != null ? ass.get(nach.getPid()) : null;

						if ((!keineInnerenSegmenteAussen && (nachSegment != null || vonSegment != null))
								|| (nachSegment != null && vonSegment != null)) {
							final InneresStrassenSegment segment = new InneresStrassenSegment(obj, laenge);

							if (nachSegment != null) {
								nachSegment.addVorgaenger(segment);
								segment.addNachfolger(nachSegment);
							}
							if (vonSegment != null) {
								vonSegment.addNachfolger(segment);
								segment.addVorgaenger(vonSegment);
							}

							if (nachSegment != null && vonSegment != null
									&& nachSegment.liegtAufSelberStrasse(vonSegment)) {
								segment.addVorgaengerAufStrasse(vonSegment);
								vonSegment.addNachfolgerAufStrasse(segment);
								segment.addNachfolgerAufStrasse(nachSegment);
								nachSegment.addVorgaengerAufStrasse(segment);
							}

							final StrassenKnoten strassenKnoten;
							if (nachSegment != null) {
								strassenKnoten = nachSegment.getVonKnoten();
							} else {
								strassenKnoten = vonSegment.getNachKnoten();
							}
							strassenKnoten.addInneresSegment(segment);

							result.put(obj.getPid(), segment);
						}
					} else {
						_debug.warning("Daten von " + atgISS + " nicht versorgt für " + obj
								+ "\n=> KANN NICHT VERWENDET WERDEN");
					}
				} else {
					_debug.warning(
							"Daten von " + atgSS + " nicht versorgt für " + obj + "\n=> KANN NICHT VERWENDET WERDEN");
				}
			}
		}

		return result;
	}

	/**
	 * Bestimmt alle Messquerschnitte auf einer Menge von Strassensegmenten.
	 * 
	 * @param connection
	 *            DAV-Verbindung
	 * @param strassenSegmente
	 *            Sammlung von Straßensegmente, deren MQ bestimmt werden sollen
	 * @return Sammlung aller gefundenen MQ
	 */
	private Map<String, MessQuerschnitt> bestimmeMessQuerschnitte(final ClientDavInterface connection,
			final Map<String, ? extends StrassenSegment> strassenSegmente) {
		final Map<String, MessQuerschnitt> result = new HashMap<String, MessQuerschnitt>();
		final SystemObjectType type = connection.getDataModel().getType("typ.messQuerschnittAllgemein");
		final AttributeGroup atg = connection.getDataModel().getAttributeGroup("atg.punktLiegtAufLinienObjekt");

		for (SystemObject obj : type.getElements()) {
			if (obj.isValid()) {
				final Data data = obj.getConfigurationData(atg);
				if (data != null) {
					final ReferenceValue linienRef = data.getItem("LinienReferenz").asReferenceValue();
					final StrassenSegment segment = strassenSegmente.get(linienRef.getSystemObject().getPid());
					if (segment != null) {
						final long offset = data.getItem("Offset").asUnscaledValue().longValue() / 100;
						final MessQuerschnitt mq = new MessQuerschnitt(obj, segment, offset);
						result.put(obj.getPid(), mq);
						segment.addMessQuerschnitt(mq);
					}
				}
			}
		}

		return result;
	}

	/**
	 * Bestimmt alle Strassenteilsegmente einer Menge von Strassensegmenten.
	 * 
	 * @param connection
	 *            DAV-Verbindung
	 * @param strassenSegmente
	 *            Sammlung von Straßensegmente, deren Strassenteilsegmente
	 *            bestimmt werden sollen
	 * @return Sammlung aller gefundenen Strassenteilsegmente
	 */
	private Map<String, StrassenTeilSegment> bestimmeStrassenTeilSegmente(final ClientDavInterface connection,
			final Map<String, ? extends StrassenSegment> strassenSegmente) {
		final Map<String, StrassenTeilSegment> result = new HashMap<String, StrassenTeilSegment>();
		final AttributeGroup atg = connection.getDataModel().getAttributeGroup("atg.bestehtAusLinienObjekten");

		for (StrassenSegment segment : strassenSegmente.values()) {
			final Data data = segment.getDavObjekt().getConfigurationData(atg);

			if (data != null) {
				final AttributeGroup atgSts = connection.getDataModel().getAttributeGroup("atg.straßenTeilSegment");
				final ReferenceArray array = data.getItem("LinienReferenz").asReferenceArray();
				long offset = 0;

				for (ReferenceValue ref : array.getReferenceValues()) {
					final SystemObject val = ref.getSystemObject();
					if (val.isOfType("typ.straßenTeilSegment")) {
						final Data dataSts = val.getConfigurationData(atgSts);
						if (dataSts != null) {
							final long laenge = dataSts.getItem("Länge").asUnscaledValue().longValue() / 100;
							final int anzahlFahrStreifen = dataSts.getItem("AnzahlFahrStreifen").asUnscaledValue()
									.intValue();
							final String steigungGefaelle = dataSts.getItem("SteigungGefälle").asTextValue().getText();

							final StrassenTeilSegment sts = new StrassenTeilSegment(val, laenge, anzahlFahrStreifen,
									steigungGefaelle, offset);
							segment.addStrassenTeilSegment(sts);
							offset += laenge;
							result.put(sts.getDavObjekt().getPid(), sts);
						}
					}
				}
			}
		}

		return result;
	}

	/**
	 * Bestimmt alle Straßenabschnitte, die durch eine Menge von MQs gegeben.
	 * sind
	 * 
	 * @param connection
	 *            DAV-Verbindung
	 * @param mq
	 *            Sammlung von MQ, die die zu bestimmenden Straßenabschnitte
	 *            vorgeben
	 * @return Sammlung aller gefundenen Straßenabschnitte
	 */
	private Map<String, StrassenAbschnitt> bestimmeStrassenAbschnitte(final ClientDavInterface connection,
			final Map<String, MessQuerschnitt> mq) {
		final Map<String, StrassenAbschnitt> result = new HashMap<String, StrassenAbschnitt>();

		final SystemObjectType type = connection.getDataModel().getType("typ.straßenAbschnitt");
		final AttributeGroup atg = connection.getDataModel().getAttributeGroup("atg.straßenAbschnitt");

		for (SystemObject obj : type.getElements()) {
			if (obj.isValid()) {
				final Data data = obj.getConfigurationData(atg);
				if (data != null) {
					final SystemObject von = data.getItem("vonMessQuerschnitt").asReferenceValue().getSystemObject();
					final MessQuerschnitt vonMq = mq.get(von.getPid());
					if (vonMq != null) {
						final SystemObject bis = data.getItem("bisMessQuerschnitt").asReferenceValue()
								.getSystemObject();
						final MessQuerschnitt bisMq = mq.get(bis.getPid());
						if (bisMq != null) {
							result.put(obj.getPid(), new StrassenAbschnitt(obj, vonMq, bisMq));
						}
					}
				}
			}
		}

		return result;
	}

}
