/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.meldungserstellung;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.tmc.allgemein.RDSEreignis;
import de.bsvrz.ste.tmc.konfiguration.AeusseresStrassenSegment;
import de.bsvrz.ste.tmc.konfiguration.InneresStrassenSegment;
import de.bsvrz.ste.tmc.konfiguration.Konfiguration;
import de.bsvrz.ste.tmc.konfiguration.StrassenSegment;
import de.bsvrz.ste.tmc.konfiguration.TmcLocationPunkt;
import de.bsvrz.ste.tmc.parameter.RDSParameterDaten;
import de.bsvrz.ste.tmc.parameter.RDSParameterListener;
import de.bsvrz.ste.tmc.parameter.RDSParameterVerwaltung;
import de.bsvrz.ste.tmc.situation.Situation;
import de.bsvrz.ste.tmc.situation.SituationEigenschaften;
import de.bsvrz.ste.tmc.situation.SituationEigenschaftenListener;
import de.bsvrz.ste.tmc.situation.VerwaltungBaustellen;
import de.bsvrz.ste.tmc.verkehrsanalyse.SegmentSituationDaten;
import de.bsvrz.ste.tmc.verkehrsanalyse.StoerfallSituation;
import de.bsvrz.ste.tmc.verkehrsanalyse.StoerfallSituationDaten;
import de.bsvrz.ste.tmc.verkehrsanalyse.Verkehrsanalyse;
import de.bsvrz.sys.funclib.concurrent.UnboundedQueue;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * 
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class Meldungserstellung
		implements RDSParameterListener, EreignisListener, Runnable, SituationEigenschaftenListener {
	/**
	 * Debug-Logger für Logging-Ausgaben
	 */
	private final static Debug _debug = Debug.getLogger();

	LocationMethode _locationMethode;

	LocationMethode getLocationMethode() {
		return _locationMethode;
	}

	PublikationInitial _publikationInitial;

	PublikationInitial getPublikationInitial() {
		return _publikationInitial;
	}

	/**
	 * Mutex für Ereignis-Queue
	 */
	public final Object mutexEreignis = new Object();

	/**
	 * Die Verwaltung der Konfigurationsobjekte des Netztes
	 */
	private Konfiguration _konfiguration;

	// public Konfiguration getKonfiguration()
	// {
	// return _konfiguration;
	// }

	/**
	 * Die Routen, die von dieser Instanz verwaltet werden
	 */
	private Collection<Route> _routen = new ArrayList<Route>();

	/**
	 * @return die Routen, die von dieser Instanz verwaltet werden - Read only
	 */
	public Collection<Route> getRouten() {
		return Collections.unmodifiableCollection(_routen);
	}

	// ClientDavInterface getConnection()
	// {
	// return getKonfiguration().getConnection();
	// }

	ClientDavInterface _connection;

	private VerwaltungMeldungen _verwaltungMeldungen;

	private VerwaltungBaustellen _baustellenVerwaltung;

	public Meldungserstellung(final ClientDavInterface connection, final Konfiguration konfiguration,
			final RDSParameterVerwaltung parameterVerwaltung, final MeldungenGlobal meldungenGlobal,
			final boolean alleSegmente) {
		_connection = connection;
		_konfiguration = konfiguration;

		parameterVerwaltung.addParameterListener(this);
		neueParameter(parameterVerwaltung, parameterVerwaltung.getAktuelleDaten());

		bestimmeRouten();
		bestimmeNullEreignisse();

		_verwaltungMeldungen = new VerwaltungMeldungen(connection, parameterVerwaltung.getLms(), konfiguration,
				meldungenGlobal);
		_ereignisse.putAll(_verwaltungMeldungen.initialisiereVorhandeneMeldungen());

		_baustellenVerwaltung = new VerwaltungBaustellen(connection, konfiguration.getNetz(),
				new SituationEigenschaftenListener[] { this });

		_verkehrsanalyse = new Verkehrsanalyse(connection, konfiguration, parameterVerwaltung, getRouten(),
				alleSegmente);

		Thread thread = new Thread(this);
		thread.setName("RDS-Ereignis Thread");
		thread.start();
		_debug.config(thread + " gestartet");
	}

	private Map<Route, Map<String, Ereignis>> _nullEreignisse = new HashMap<Route, Map<String, Ereignis>>();

	private void bestimmeNullEreignisse() {
		for (Route route : getRouten()) {
			Map<String, Ereignis> map = new LinkedHashMap<String, Ereignis>();
			_nullEreignisse.put(route, map);

			for (StrassenSegment segment : route.getSegmente()) {
				if (segment instanceof AeusseresStrassenSegment) {
					TmcLocationPunkt tmcOrt = ((AeusseresStrassenSegment) segment).getTmcPunkt();
					int tmcRichtung = ((AeusseresStrassenSegment) segment).getTmcRichtung();
					ArrayList<TmcLocationPunkt> orte = new ArrayList<TmcLocationPunkt>();
					orte.add(tmcOrt);
					Ereignis nullEreignis = new Ereignis(Ereignis.Art.NULL, RDSEreignis.KEINE_AUSSAGE, orte,
							tmcRichtung, null);
					map.put(nullEreignis.getSchluessel(), nullEreignis);
				}
			}
		}
	}

	private Map<StrassenSegment, Set<Route>> _routenZuSegment = new HashMap<StrassenSegment, Set<Route>>();

	Collection<Route> getRoutenZuSegment(final StrassenSegment segment) {
		Collection<Route> col = _routenZuSegment.get(segment);

		if (col == null) {
			return Collections.emptySet();
		} else {
			return Collections.unmodifiableCollection(col);
		}
	}

	private void bestimmeRouten() {
		// Bestimme Routen
		Collection<List<StrassenSegment>> listen = erzeugeRouten();
		String text = "Routenverläufe:";
		for (List<StrassenSegment> liste : listen) {
			Route route = new Route(this, liste);
			_routen.add(route);
			route.addEreignisListener(this);
			text += "\n";
			for (StrassenSegment segment : liste) {
				Set<Route> routen = _routenZuSegment.get(segment);
				if (routen == null) {
					routen = new HashSet<Route>();
					_routenZuSegment.put(segment, routen);
				}
				routen.add(route);

				text += segment.getDavObjekt().getPid();
				// Ausgabe für Routenverlauf
				if (segment instanceof AeusseresStrassenSegment && segment.endetInAutobahnKreuzOderDreieck()) {
					text += "*";
				} else {
					text += " ";
				}
				if (!segment.getNachfolgerAufStrasse().isEmpty()) {
					text += "=>";
				}
			}
		}
		_debug.finer(text);
	}

	private static Collection<List<StrassenSegment>> berechneRouten(final LinkedList<StrassenSegment> route) {
		Set<List<StrassenSegment>> result = new HashSet<List<StrassenSegment>>();
		StrassenSegment letztes = route.getLast();
		Collection<StrassenSegment> nachfolger = letztes.getNachfolgerAufStrasse();

		if (!nachfolger.isEmpty()) {
			if (false /* Behandlunge aller Nachfolger - z. Z. nicht verwendet */) {
				for (StrassenSegment segment : nachfolger) {
					LinkedList<StrassenSegment> liste = new LinkedList<StrassenSegment>(route);
					liste.add(segment);
					result.addAll(berechneRouten(liste));
				}
			} else /* Behandle nur kürzesten Nachfolger auf Strasse */
			{
				LinkedList<StrassenSegment> liste = new LinkedList<StrassenSegment>(route);
				StrassenSegment segment = letztes.getKuerzestenNachfolgerAufStrasse();

				if (segment != null) {
					liste.add(segment);
					result.addAll(berechneRouten(liste));
				}
			}
		} else {
			result.add(new LinkedList<StrassenSegment>(route));
		}
		return result;
	}

	private Collection<List<StrassenSegment>> erzeugeRouten() {
		_debug.fine("Berechne Routen");

		// Finde alle äusseren Strassensegmente ohne Vorgänger
		List<LinkedList<StrassenSegment>> anfangsSegmente = new ArrayList<LinkedList<StrassenSegment>>();

		for (AeusseresStrassenSegment segment : _konfiguration.getAeussereStrassenSegmente().values()) {
			if (segment.getVorgaengerAufStrasse().isEmpty()) {
				LinkedList<StrassenSegment> liste = new LinkedList<StrassenSegment>();
				liste.add(segment);
				anfangsSegmente.add(liste);
			}
		}
		_debug.fine(anfangsSegmente.size() + " Anfangssegmente für Routenberechnung");

		// Bestimme Routen
		Set<List<StrassenSegment>> routen = new HashSet<List<StrassenSegment>>();
		for (LinkedList<StrassenSegment> root : anfangsSegmente) {
			routen.addAll(berechneRouten(root));
		}
		_debug.config(routen.size() + " Routen berechnet");

		return routen;
	}

	/**
	 * Verkehrsanalyse-Instanz
	 */
	private Verkehrsanalyse _verkehrsanalyse;

	public Verkehrsanalyse getVerkehrsanalyse() {
		return _verkehrsanalyse;
	}

	private RDSParameterDaten _parameter;

	public RDSParameterDaten getParameter() {
		return _parameter;
	}

	public void neueParameter(final RDSParameterVerwaltung quelle, final RDSParameterDaten daten) {
		byte level = daten.getLevelFlags(getParameter());

		synchronized (mutexEreignis) {
			_parameter = daten;

			if ((level & RDSParameterDaten.FLAG_LEVEL1) != 0) {
				_debug.fine("Neue Parameter: notify alle Routen");
				for (Route route : getRouten()) {
					route.fireEreignisListener();
				}
			}
		}
	}

	UnboundedQueue<Ereignis> _queue = new UnboundedQueue<Ereignis>();

	private Map<AeusseresStrassenSegment, List<EreignisDaten>> zuPrimaereTmcLocation(final List<EreignisDaten> daten) {
		Map<AeusseresStrassenSegment, List<EreignisDaten>> result = new LinkedHashMap<AeusseresStrassenSegment, List<EreignisDaten>>();

		for (EreignisDaten ereignisDaten : daten) {
			AeusseresStrassenSegment tmcPrimaerSegment = null;
			if (ereignisDaten.endSegment instanceof InneresStrassenSegment) {
				tmcPrimaerSegment = ((InneresStrassenSegment) ereignisDaten.endSegment).getVonSegment();
			}
			if (ereignisDaten.endSegment instanceof AeusseresStrassenSegment) {
				tmcPrimaerSegment = (AeusseresStrassenSegment) ereignisDaten.endSegment;
			}
			List<EreignisDaten> ereignisListe = result.get(tmcPrimaerSegment);
			if (ereignisListe == null) {
				ereignisListe = new LinkedList<EreignisDaten>();
				result.put(tmcPrimaerSegment, ereignisListe);
			}
			ereignisListe.add(ereignisDaten);
		}

		return result;
	}

	public void neueEreignisse(final Route quelle, final List<EreignisDaten> daten) {
		boolean nichtAddieren = getParameter().getRDSNichtAddieren();
		int dMaxEinzelStoerfall = getParameter().getRDSDMaxEinzelStoerfall();

		Map<String, Ereignis> ereignisMap = new HashMap<String, Ereignis>();
		Map<AeusseresStrassenSegment, List<EreignisDaten>> datenMap = zuPrimaereTmcLocation(daten);

		int richtung = ((AeusseresStrassenSegment) quelle.getSegmente().get(0)).getTmcRichtung();

		for (AeusseresStrassenSegment tmcPrimaerSegment : datenMap.keySet()) {
			LinkedList<TmcLocationPunkt> tmcOrte = new LinkedList<TmcLocationPunkt>();
			List<EreignisDaten> ereignisListe = datenMap.get(tmcPrimaerSegment);
			long laengeBerechnet = 0;
			RDSEreignis ereignis = ereignisListe.get(0).ereignis;
			StrassenSegment wurzelSegment = null;
			long wurzelOffset = -1;

			_debug.finer("Längenberechnung: ");

			// Primäre Location
			tmcOrte.add(tmcPrimaerSegment.getTmcPunkt());

			for (EreignisDaten ereignisDaten : ereignisListe) {
				// Aufbau der Liste der TMC-Locations für eine RDS-Meldung
				// Weitere Locations, wenn vorhanden
				AeusseresStrassenSegment as = tmcPrimaerSegment;
				while (as != ereignisDaten.startSegment) {
					InneresStrassenSegment is = (InneresStrassenSegment) as.getKuerzestenVorgaengerAufStrasse();
					as = is != null ? is.getVonSegment() : null;
					if (as == null) {
						break;
					}
					if (!tmcOrte.contains(as.getTmcPunkt())) {
						tmcOrte.add(as.getTmcPunkt());
					}
				}
				_debug.finer("addiere Länge: " + ereignisDaten.getLaenge() + "(" + ereignisDaten.toString() + ")");
				laengeBerechnet += ereignisDaten.getLaenge();

				wurzelSegment = ereignisDaten.endSegment;
				wurzelOffset = ereignisDaten.ende;
			}
			// Füge letzte (sekundäre) Location hinzu
			tmcOrte.add(tmcOrte.getLast().getVorgaenger());

			if ((nichtAddieren && ereignisListe.size() > 1) || // Falls
																// Summierung
																// der Längen
																// von
																// Einzelereignissen
																// nicht
																// gewünscht
					(ereignisListe.size() == 1 && ereignisListe.get(0).lokal != null && // oder
																						// falls
																						// einzelner
																						// lokaler
																						// Stoerfall
																						// und
							laengeBerechnet > dMaxEinzelStoerfall)) // Länge zu
																	// gross
			{
				laengeBerechnet = -1;
			}

			String code = null;
			RDSQuantitaetLEN rdsQuantitaetLEN = null;
			final SystemObject baustelle = getErklaerungsKomponente(wurzelSegment, wurzelOffset);

			if (laengeBerechnet == -1) {
				// Von vorneherein keine Längenangabe erwünscht => hole
				// Ereiniscode ohne Längenangabe
				code = RDSBeschreibung.getCodeZuLaenge(ereignis, RDSBeschreibung.KEINE_LAENGE, null, baustelle != null);
			} else {
				final int laengeInGanzenKm = RDSQuantitaetLEN.getLaengeInGanzenKm(laengeBerechnet);
				final Collection<String> zulaessigeCodes = getParameter().getRDSAbbildungCode().get(ereignis);

				_debug.finer("Ereigniscode wird bestimmt aus: Länge(berechnet): " + laengeBerechnet + ", Länge(ganz): "
						+ laengeInGanzenKm + ", Codeliste: " + zulaessigeCodes);

				if (zulaessigeCodes != null) {
					code = RDSBeschreibung.getCodeZuLaenge(ereignis, laengeInGanzenKm, zulaessigeCodes,
							baustelle != null);
				}

				_debug.finer("bestimmeter Ereigniscode: " + code);

				if (code == null) {
					// Ereigniscode konnte aus Ereigniskatalog nicht bestimmt
					// werden
					// Hole Ereiniscode ohne Längenangabe:
					code = RDSBeschreibung.getCodeZuLaenge(ereignis, RDSBeschreibung.KEINE_LAENGE, null,
							baustelle != null);
					// Zur Längenangabe wird rdsQuantitaetLEN verwendet
					rdsQuantitaetLEN = new RDSQuantitaetLEN(laengeInGanzenKm);
				}
				// else
				// {
				// Ereigniscode konnte aus Ereigniskatalog bestimmt werden
				// rdsQuantitaetLEN wird nicht verwendet
				// }
			}

			final Ereignis event = new Ereignis(Ereignis.Art.NORMAL, ereignis, tmcOrte, richtung, code);
			event.setRDSQuantitaetLEN(rdsQuantitaetLEN);
			event.setDatenListe(ereignisListe);
			event.setErklaerungsKomponente(baustelle);
			ereignisMap.put(event.getSchluessel(), event);
			// for (String ort: orte)
			// {
			// notNullSet.add(ort);
			// }
		}

		for (String key : _nullEreignisse.get(quelle).keySet()) {
			final Ereignis effektivEreignis = ereignisMap.get(key);
			if (effektivEreignis != null) {
				_queue.put(effektivEreignis);
			} else {
				final Ereignis nullEreignis = _nullEreignisse.get(quelle).get(key);
				_queue.put(nullEreignis);
				// if (!notNullSet.contains(nullEreignis.getTmcPrimaererOrt()))
				// {
				// _queue.put(nullEreignis);
				// }
			}
		}
	}

	public final static long MIN_TIMEOUT = 500;
	public final static long MAX_TIMEOUT = 30000;

	private Map<String, Ereignis> _ereignisse = new LinkedHashMap<String, Ereignis>();

	private void ausgabeDaten() {
		for (Route route : getRouten()) {
			int anzahl = 0;

			for (StrassenSegment segment : route.getSegmente()) {
				StoerfallSituation stoerfallSituation = _verkehrsanalyse
						.getStoerfallSituation(segment.getDavObjekt().getPid());

				if (stoerfallSituation != null) {
					anzahl += 1;
					AeusseresStrassenSegment tmcSegment = segment instanceof AeusseresStrassenSegment
							? (AeusseresStrassenSegment) segment : ((InneresStrassenSegment) segment).getVonSegment();
					String tmcInfo = tmcSegment.getTmcPunkt().getLocationCode() + " "
							+ (tmcSegment.getTmcRichtung() < 0 ? "-1" : "+1");

					String text;

					// text = "Zustände aller SI nach Überlagerung im Segment "
					// + segment + " (TMC-Location " + tmcInfo + ":";
					// for (StoerfallDaten daten:
					// stoerfallSituation.getUeberlagerteDaten())
					// {
					// text += "\n\"" + daten.zustand + "\" - " + daten.start +
					// "m bis " + daten.ende +
					// "m von " + daten.stoerfall.getDavObjekt().getPid() + "
					// mit Priorität " + daten.prioritaet;
					// }
					// _debug.fine(text);

					text = "Jeweils schärfster Zustand nach Überlagerung im Segment " + segment + " (TMC-Location "
							+ tmcInfo + ":";
					List<StoerfallSituationDaten> liste = route.getStoerfalleAufSegment(segment);
					if (liste != null && !liste.isEmpty()) {
						for (StoerfallSituationDaten daten : liste) {
							text += "\n\"" + daten.zustand + "\" - " + daten.start + "m bis " + daten.ende + "m";
						}
					} else {
						text += "\n\t<KEINE>";
					}
					_debug.fine(text);
				}
			}

			String text = "Für die vorstehenden " + anzahl + " aufeinanderfolgenden TMC-Abschnitte "
					+ "\nwerden die folgenden Daten zur Meldungsermittlung benutzt:";
			for (EreignisDaten daten : route.getEreignisDaten()) {
				text += "\n\"" + daten.ereignis + "\" auf einer Länge von " + daten.getLaenge() + " beginnend bei "
						+ daten.endSegment.getDavObjekt().getPid() + " + " + daten.start + "m";
			}
			_debug.fine(text);
		}
	}

	private void bearbeiteEreignisse(final Map<String, Ereignis> ereignisse) {
		final Map<String, Ereignis> neueDaten = new LinkedHashMap<String, Ereignis>();

		_debug.fine("Ereignisse werden bearbeitet");

		synchronized (_ereignisse) {

			for (String schluessel : ereignisse.keySet()) {
				final Ereignis soll = ereignisse.get(schluessel);
				final Ereignis ist = _ereignisse.get(schluessel);

				_debug.fine("Soll: " + soll + ", "
						+ ((soll != null) && (_verwaltungMeldungen.getMeldung(schluessel) != null)
								? _verwaltungMeldungen.getMeldung(schluessel).getDavObjekt().getPid() : "<>"));
				_debug.fine(
						"Ist:  " + ist + ", " + ((ist != null) && (_verwaltungMeldungen.getMeldung(schluessel) != null)
								? _verwaltungMeldungen.getMeldung(schluessel).getDavObjekt().getPid() : "<>"));

				if (soll == null) {
					continue;
				}

				if (!soll.equals(ist)) {
					// if (ist != null || soll.getRDSEreignis() !=
					// RDSEreignis.KEINE_AUSSAGE)
					// {
					// _debug.config("Neues RDS-Ereignis:\n" + soll);
					// }

					synchronized (_verwaltungMeldungen) {
						final Meldung meldung = _verwaltungMeldungen.getMeldung(schluessel);
						if (meldung != null && ist != null && ist.getRDSEreignis() != RDSEreignis.KEINE_AUSSAGE) {
							if (soll.getRDSEreignis() != RDSEreignis.KEINE_AUSSAGE) {
								// bedingteAufhebung(soll, neueDaten);
								_debug.config(soll.getDatenBeschreibung());

								_debug.info("Die bestehende TMC-Meldung " + meldung.getDavObjekt().getPid()
										+ " wird geändert mit:\n" + soll);
								neueDaten.put(schluessel, soll);
							} else {
								_debug.info("Die bestehende TMC-Meldung " + meldung.getDavObjekt().getPid() + " mit\n"
										+ ist + "\nwird aufgehoben");
								neueDaten.put(schluessel,
										new Ereignis(Ereignis.Art.NORMAL, RDSEreignis.KEINE_AUSSAGE, ist));
							}
						} else {
							if (soll.getRDSEreignis() != RDSEreignis.KEINE_AUSSAGE) {
								// bedingteAufhebung(soll, neueDaten);
								_debug.config(soll.getDatenBeschreibung());

								_debug.info("Eine neue TMC-Meldung wird erstellt für:\n" + soll);
								if (_verwaltungMeldungen.erzeugeMeldung(schluessel)) {
									neueDaten.put(schluessel, soll);
								}
							}
						}
					}
				}
			}
			// _ereignisse.putAll(neueDaten);
			for (String schluessel : neueDaten.keySet()) {
				final Ereignis ereignis = neueDaten.get(schluessel);
				final Meldung meldung = _verwaltungMeldungen.getMeldung(schluessel);

				// ResultData resultData =
				// _verwaltungMeldungen.getData(meldung.getDavObjekt(),
				// ereignis);
				if (meldung.send(ereignis)) {
					_debug.fine("Senden erfolgreich");
					_ereignisse.put(schluessel, ereignis);
				}
			}
		}
	}

	// private void bedingteAufhebung(Ereignis ereignis, Map<String, Ereignis>
	// daten)
	// {
	// if (ereignis.getTmcOrte().size() > 1)
	// {
	// for (String schluessel: _ereignisse.keySet())
	// {
	// Meldung meldung = _meldungen.get(schluessel);
	// if (meldung != null)
	// {
	// Ereignis lauf = _ereignisse.get(schluessel);
	// if (lauf != null && lauf.getRDSEreignis() != RDSEreignis.KEINE_AUSSAGE &&
	// lauf.getTmcRichtung() == ereignis.getTmcRichtung())
	// {
	// for (int i = 1; i < ereignis.getTmcOrte().size(); i+=1)
	// {
	// if (lauf.getTmcPrimaererOrt().equals(ereignis.getTmcOrte().get(i)))
	// {
	// _debug.config("Die bestehende RDS-Meldung " +
	// meldung.getDavObjekt().getPid() +
	// " mit\n" + lauf + "\nwird aufgehoben");
	// daten.put(schluessel, new Ereignis(RDSEreignis.KEINE_AUSSAGE, lauf));
	// }
	// }
	// }
	// }
	// }
	// }
	// }

	/*
	 * (Kein Javadoc)
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		Map<String, Ereignis> daten = new LinkedHashMap<String, Ereignis>();

		long startZeit = System.currentTimeMillis();

		while (true) {
			try {
				final Ereignis ereignis = _queue.take();

				_debug.fine("Ereignis: " + ereignis);

				daten.put(ereignis.getSchluessel(), ereignis);

				final boolean leer;

				synchronized (_queue) {
					if (_queue.size() == 0) {
						_queue.wait(MIN_TIMEOUT);
					}
					leer = _queue.size() == 0;
				}

				if (leer || System.currentTimeMillis() - MAX_TIMEOUT > startZeit) {
					bearbeiteEreignisse(daten);
					daten.clear();
					startZeit = System.currentTimeMillis();
				}

			} catch (InterruptedException e) {
				_debug.warning("Thread wurde unerwartet unterbrochen", e);
			}
		}

	}

	private Map<StrassenSegment, List<SegmentSituationDaten>> _segmentBaustellenDaten = new HashMap<StrassenSegment, List<SegmentSituationDaten>>();

	Collection<SegmentSituationDaten> getBaustellenDaten(final StrassenSegment segment) {
		Collection<SegmentSituationDaten> col = _segmentBaustellenDaten.get(segment);
		if (col == null) {
			return Collections.emptySet();
		} else {
			return Collections.unmodifiableCollection(col);
		}
	}

	public void neueSituationEigenschaften(final Situation quelle, final SituationEigenschaften daten) {
		_debug.fine("Neue Baustellendaten für " + quelle + ": "
				+ (daten == null ? "<KEINE>"
						: "start = " + daten.startOffset + ", ende = " + daten.endOffset + ", "
								+ daten.strassenSegment.length + " Segmente"));

		synchronized (_segmentBaustellenDaten) {
			// Entferne zunächst alle vorhanden Daten für diese Situation
			Iterator<StrassenSegment> outer = _segmentBaustellenDaten.keySet().iterator();
			while (outer.hasNext()) {
				StrassenSegment segment = outer.next();
				List<SegmentSituationDaten> liste = _segmentBaustellenDaten.get(segment);

				Iterator<SegmentSituationDaten> inner = liste.iterator();
				while (inner.hasNext()) {
					SegmentSituationDaten segmentDaten = inner.next();
					if (segmentDaten.situation == quelle) {
						inner.remove();
					}
				}
				if (liste.isEmpty()) {
					outer.remove();
				}
			}

			// Füge neue Daten hinzu
			if (daten != null && daten.istGueltig(_connection.getTime())) {
				for (int i = 0; i < daten.strassenSegment.length; i += 1) {
					SystemObject objekt = daten.strassenSegment[i];
					StrassenSegment segment = _konfiguration.getAeussereStrassenSegmente().get(objekt.getPid());
					if (segment == null) {
						segment = _konfiguration.getInnereStrassenSegmente().get(objekt.getPid());
					}
					if (segment != null) {
						long start = i == 0 ? daten.startOffset : 0;
						long ende = i == daten.strassenSegment.length - 1 ? daten.endOffset : segment.getLaenge();

						List<SegmentSituationDaten> liste = _segmentBaustellenDaten.get(segment);
						if (liste == null) {
							liste = new LinkedList<SegmentSituationDaten>();
							_segmentBaustellenDaten.put(segment, liste);
						}
						liste.add(new SegmentSituationDaten(quelle, segment, start, ende));
					}
				}
			}
		}

		synchronized (mutexEreignis) {
			if (getParameter().getRDSBeruecksichtigeBaustellen()) {
				for (Route route : getRouten()) {
					route.fireEreignisListener();
				}
			}
		}

	}

	private Situation existiertBaustelle(final StrassenSegment startSegment, final long startOffset,
			final StrassenSegment endSegment, final long endOffset) {
		StrassenSegment segment = endSegment;

		while (segment != null) {
			// Setze Start/-Ende-Offsets
			long start = segment == startSegment ? startOffset : 0;
			long ende = segment == endSegment ? endOffset : segment.getLaenge();

			synchronized (_segmentBaustellenDaten) {
				Collection<SegmentSituationDaten> baustellenListe = _segmentBaustellenDaten.get(segment);

				if (baustellenListe != null) {
					// Iteriere Stau-Anteile für das aktuelle Segment, Baustelle
					// existiert bei Überlappung
					for (SegmentSituationDaten baustellenDaten : baustellenListe) {
						if ((start >= baustellenDaten.startOffset && start < baustellenDaten.endOffset)
								|| (ende > baustellenDaten.startOffset && ende <= baustellenDaten.endOffset)
								|| (start < baustellenDaten.startOffset && ende > baustellenDaten.endOffset)) {

							return baustellenDaten.situation;
						}
					}
				}
			}

			// Hole nächstes Strassensegment wenn möglich
			if (segment == startSegment) {
				segment = null;
			} else {
				segment = segment.getKuerzestenVorgaengerAufStrasse();
			}
		}
		return null;
	}

	private SystemObject getErklaerungsKomponente(final StrassenSegment wurzelSegment, final long wurzelOffset) {
		SystemObject result = null;

		if (getParameter().getRDSBeruecksichtigeBaustellen()) {
			long abstandVor = getParameter().getRDSDBaustelleSuchbereichNach();
			StrassenSegment startSegment = wurzelSegment;
			long startOffset = wurzelOffset - abstandVor;

			while (startOffset < 0) {
				StrassenSegment segment = startSegment.getKuerzestenVorgaengerAufStrasse();
				if (segment != null) {
					startOffset += startSegment.getLaenge();
					startSegment = segment;
				} else {
					startOffset = 0;
				}
			}

			long abstandNach = getParameter().getRDSDBaustelleSuchbereichVor();
			StrassenSegment endSegment = wurzelSegment;
			long endOffset = wurzelOffset + abstandNach;

			while (endOffset > endSegment.getLaenge()) {
				StrassenSegment segment = endSegment.getKuerzestenNachfolgerAufStrasse();
				if (segment != null) {
					endOffset -= endSegment.getLaenge();
					endSegment = segment;
				} else {
					endOffset = endSegment.getLaenge();
				}
			}

			Situation baustelle = existiertBaustelle(startSegment, startOffset, endSegment, endOffset);
			return baustelle != null ? baustelle.getDavObjekt() : null;
		}

		return result;
	}

}
