/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.meldungserstellung;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.InvalidationListener;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.tmc.allgemein.RDSEreignis;
import de.bsvrz.ste.tmc.konfiguration.Konfiguration;
import de.bsvrz.ste.tmc.konfiguration.TmcLocationPunkt;
import de.bsvrz.ste.tmc.parameter.ParameterEmpfaenger;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.dynobj.DynObjektException;
import de.bsvrz.sys.funclib.dynobj.DynamischeObjekte;

/**
 * Verwaltung der Meldungen.
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class VerwaltungMeldungen implements InvalidationListener, MeldungenGlobal {
	/**
	 * Debug-Logger für Logging-Ausgaben.
	 */
	private static final Debug _debug = Debug.getLogger();

	/** Klasse zum Verwalten dynamischer Objekte. */
	private DynamischeObjekte vdo;

	private ClientDavInterface connection;

	public ClientDavInterface getConnection() {
		return connection;
	}

	private ConfigurationObject _konfigurationsObjekt;

	public ConfigurationObject getKonfigurationsObjekt() {
		return _konfigurationsObjekt;
	}

	private Konfiguration _konfiguration;

	/**
	 * Datenbeschreibung "generiert".
	 */
	private DataDescription _datenBeschreibung;

	public DataDescription getDatenBeschreibung() {
		return _datenBeschreibung;
	}

	private MeldungenGlobal _meldungenGlobal;

	/***************************************************************************/
	/* Implementierung Interface MeldungenGlobal */
	/***************************************************************************/

	/**
	 * @return den Konfigurationsbereich, in dem dynamische Meldungsobjekte
	 *         abgelegt werden
	 */
	public ConfigurationArea getKonfigurationsbereich() {
		return _meldungenGlobal.getKonfigurationsbereich();
	}

	/**
	 * @return die LocationMethode, nach der Meldungsdatensätze kodiert werden
	 */
	public LocationMethode getLocationMethode() {
		return _meldungenGlobal.getLocationMethode();
	}

	/**
	 * @return die Methode, nach der beim Programmstart aktuelle Zustände
	 *         publiziert werden
	 */
	public PublikationInitial getPublikationInitial() {
		return _meldungenGlobal.getPublikationInitial();
	}

	/**
	 * @return die (Text-) Id der Organisationseinheit, die einer generierten
	 *         Meldung zugeordnet sein soll - i. d. R. LMS
	 */
	public String getIdOrganisation() {
		return _meldungenGlobal.getIdOrganisation();
	}

	/**
	 * @return die (numerische) Id der Organisationseinheit, die einer
	 *         generierten Meldung zugeordnet sein soll - i. d. R. LMS
	 */
	public String getIdNummer() {
		return _meldungenGlobal.getIdNummer();
	}

	/**
	 * @return Zeitspanne in min., wielange eine Meldung gültig (nicht
	 *         abgelaufen) sein soll, nachdem sie aktualisiert wurde
	 */
	public int getAblaufOffset() {
		return _meldungenGlobal.getAblaufOffset();
	}

	/***************************************************************************/
	/* Konstanten für die Meldungsverwaltung */
	/***************************************************************************/

	public static final String MENGE = "TMCMeldungen";

	public static final String PID_PREFIX = "tmcMeldung";

	public static final String NAME_PREFIX = "TMCMeldung";

	public VerwaltungMeldungen(final ClientDavInterface connection, final ConfigurationObject konfigurationsObjekt,
			final Konfiguration konfiguration, final MeldungenGlobal meldungenGlobal) {
		this.connection = connection;
		_konfigurationsObjekt = konfigurationsObjekt;
		_konfiguration = konfiguration;
		_meldungenGlobal = meldungenGlobal;
		vdo = DynamischeObjekte.getInstanz(connection);

		final AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.tmcVerkehrsMeldung");
		final Aspect asp = getConnection().getDataModel().getAspect("asp.tmcGeneriert");
		_datenBeschreibung = new DataDescription(atg, asp);
	}

	private Map<String, Meldung> _meldungen = new LinkedHashMap<String, Meldung>();

	public Meldung getMeldung(final String schluessel) {
		synchronized (this) {
			return _meldungen.get(schluessel);
		}
	}

	private Ereignis bestimmeEreignis(final ResultData resultData) {
		final Data data = resultData.getData();
		final List<TmcLocationPunkt> tmcLocationen = new LinkedList<TmcLocationPunkt>();

		switch (getLocationMethode()) {
		case LISTE: {
			final List<String> locationCodes = getAlleTmcLocationen(data);
			for (String code : locationCodes) {
				tmcLocationen.add(_konfiguration.getTmcPunkteMitLocationCode().get(code));
			}
			break;
		}
		case EXTEND: {
			TmcLocationPunkt tmcPunkt = _konfiguration.getTmcPunkteMitLocationCode().get(getPrimaereTmcLocation(data));
			final int extend = getLocationExtend(data);
			while (tmcLocationen.size() <= extend) {
				tmcLocationen.add(tmcPunkt);
				tmcPunkt = tmcPunkt.getVorgaenger();
			}
			break;
		}
		default: {
			throw new UnsupportedOperationException(
					"Location-Methode mit Id \"" + getLocationMethode().id + "\" nicht unterstützt");
		}
		}

		final int tmcRichtung = getTmcRichtung(data);
		final String ereignisCode = getEreignisCode(data);
		final RDSEreignis rdsEreignis = RDSBeschreibung.getRDSEreignisZuCode(ereignisCode);

		final Ereignis ereignis = new Ereignis(Ereignis.Art.INITIAL, rdsEreignis, tmcLocationen, tmcRichtung,
				ereignisCode);
		ereignis.setRDSQuantitaetLEN(null);

		return ereignis;
	}

	public Map<String, Ereignis> initialisiereVorhandeneMeldungen() {
		final MutableSet menge = getKonfigurationsObjekt().getMutableSet(MENGE);
		if (menge != null) {
			synchronized (this) {
				int anzahl = 0;
				final Map<String, Ereignis> result = new LinkedHashMap<String, Ereignis>();
				_meldungen.clear();

				for (SystemObject objekt : menge.getElements()) {
					if (objekt.isValid()) {
						_debug.finer("Bearbeite dynamisches Objekt " + objekt.getPid());

						final ResultData datenGeneriert = new ParameterEmpfaenger(getConnection(), objekt,
								"atg.tmcVerkehrsMeldung", _datenBeschreibung).getData();

						if (datenGeneriert != null && datenGeneriert.hasData()) {
							Meldung meldung = null;
							String schluessel = null;
							DynamicObject dynObj = null;
							try {
								final Ereignis ereignis = bestimmeEreignis(datenGeneriert);
								schluessel = ereignis.getSchluessel();
								meldung = new Meldung(this, objekt, datenGeneriert);

								if (!istAufhebung(datenGeneriert.getData())) {
									// falls Ereignis keine Aufhebung ist
									_meldungen.put(schluessel, meldung);
									result.put(schluessel, ereignis);
								}
								if (objekt instanceof DynamicObject) {
									dynObj = (DynamicObject) objekt;
									dynObj.addListenerForInvalidation(this);
								}
								anzahl += 1;
								_debug.fine("Dynamisches Objekt " + objekt.getPid() + " fehlerfrei initialisiert");

								switch (getPublikationInitial()) {
								case AENDERUNGEN:
									// tue nichts
									break;
								case ALLES:
									meldung.send(meldung.getDatenGeneriert());
									break;
								}
							} catch (final RuntimeException e) {
								if (meldung != null) {
									if (dynObj != null) {
										dynObj.removeListenerForInvalidation(this);
									}
									if (schluessel != null) {
										_meldungen.remove(schluessel);
									}
									meldung.dispose();
								}
								_debug.warning("Initialisierung fehlgeschlagen für dynamisches Objekt "
										+ objekt.getPid() + ": Objekt wird nicht verwendet", e);
							}
						} else {
							_debug.fine("Dynamisches Objekt " + objekt.getPid()
									+ " wurde nicht initialisiert, da kein Datensatz unter dem Aspekt \"TMCGeneriert\" vorhanden ist");
						}
					}
				}
				_debug.config(anzahl + " vorhandene RDS-Meldungen werden verwaltet");
				return result;
			}
		} else {
			throw new IllegalArgumentException(
					"Dynamische Menge \"" + MENGE + "\" nicht vorhanden in " + getKonfigurationsObjekt());
		}
	}

	public boolean erzeugeMeldung(final String schluessel) {
		final DynamicObjectType typ = (DynamicObjectType) getConnection().getDataModel()
				.getType("typ.tmcVerkehrsMeldung");

		final List<SystemObject> meldungen = typ.getElements();
		final Set<Integer> nummern = new HashSet<Integer>();

		for (SystemObject meldung : meldungen) {
			if (meldung.isValid()) {
				final String pid = meldung.getPid();
				if (pid.startsWith(PID_PREFIX)) {
					final String suffix = pid.substring(PID_PREFIX.length());
					Integer nummer = null;
					try {
						nummer = Integer.valueOf(suffix);
						nummern.add(nummer);
					} catch (final NumberFormatException e) {
						// suffix ist keine Zahl - Schleife wird fortgesetzt
					}
				}
			}

		}

		boolean gefunden = false;
		int ext = 1;

		do {
			while (nummern.contains(ext)) {
				ext++;
			}

			final SystemObject obj = getConnection().getDataModel().getObject(PID_PREFIX + ext);
			if (obj == null || !obj.isValid()) {
				gefunden = true;
			} else {
				ext++;
			}
		} while (!gefunden);

		DynamicObject davObjekt = null;

		try {
			final MutableSet menge = getKonfigurationsObjekt().getMutableSet(MENGE);
			davObjekt = vdo.erzeugeObjektInMenge(typ, NAME_PREFIX + ext, PID_PREFIX + ext, menge);
		} catch (final DynObjektException e) {
			_debug.error("Fehler bei Erzeugen eines dynamisches Objekts vom Typ " + typ.getPid(), e);
		}

		if (davObjekt != null) {
			synchronized (this) {
				_meldungen.put(schluessel, new Meldung(this, davObjekt));
				_debug.fine("Neue TMC-Meldung '" + davObjekt.getName() + "' ohne Fehler erzeugt");
				return true;
			}
		} else {
			return false;
		}
	}

	@Override
	public void invalidObject(final DynamicObject dynObjekt) {
		_debug.config("Objekt " + dynObjekt + " gelöscht/invalidiert => wird nicht weiter verwaltet");
		synchronized (this) {
			final Iterator<String> it = _meldungen.keySet().iterator();

			while (it.hasNext()) {
				final String schluessel = it.next();
				final Meldung meldung = _meldungen.get(schluessel);
				if (meldung.getDavObjekt() == dynObjekt) {
					meldung.dispose();
					dynObjekt.removeListenerForInvalidation(this);
					it.remove();
				}
			}
		}
	}

	private static String getEreignisCode(final Data data) {
		final Data item = data.getItem("TMCDaten");
		final Data ei = item.getItem("EreignisInformationen");
		final Array array = ei.getArray("EreignisDaten");
		if (array.getLength() > 0) {
			final Data loc = array.getItem(0);
			return "" + loc.getItem("EreignisCode").asUnscaledValue().intValue();
		}

		return "";
	}

	private static String getPrimaereTmcLocation(final Data data) {
		final Data item = data.getItem("TMCDaten");
		final Data ei = item.getItem("VerortungsInformationen");
		final Array array = ei.getArray("LokationsDaten");
		if (array.getLength() > 0) {
			final Data loc = array.getItem(0);
			return loc.getItem("PrimaerLokation").asTextValue().getText();
		}

		return "";
	}

	private static List<String> getAlleTmcLocationen(final Data data) {
		final List<String> result = new LinkedList<String>();

		final Data item = data.getItem("TMCDaten");
		final Data ei = item.getItem("VerortungsInformationen");
		final Array array = ei.getArray("LokationsDaten");
		for (int i = 0; i < array.getLength(); i++) {
			final Data loc = array.getItem(i);
			final int prim = loc.getItem("PrimaerLokation").asUnscaledValue().intValue();
			final int sek = loc.getItem("SekundaerLokation").asUnscaledValue().intValue();

			if (prim != 0) {
				result.add("" + prim);
			}

			if (sek != 0) {
				result.add("" + sek);
			}
		}

		return result;
	}

	private static int getTmcRichtung(final Data data) {
		final Data item = data.getItem("TMCDaten");
		final Data ei = item.getItem("VerortungsInformationen");
		final Array array = ei.getArray("LokationsDaten");
		if (array.getLength() > 0) {
			final Data loc = array.getItem(0);
			return loc.getItem("TMCRichtung").asUnscaledValue().intValue();
		}

		return -1;
	}

	private static int getLocationExtend(final Data data) {
		final Data dataVersion = data.getItem("Version");
		final Data dataVerkehr = dataVersion.getItem("VerkehrsInformationen");
		final Data dataLocationDaten = dataVerkehr.getItem("LocationDaten");
		return dataLocationDaten.getItem("LocationExtent").asUnscaledValue().intValue();
	}

	private static boolean istGeneriert(final Data data) {
		final Data dataVersion = data.getItem("TMCDaten");
		final Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
		return dataVerwaltung.getItem("Erzeugungsart").asTextValue().getText().equals("automatisch");
	}

	private static boolean istAufhebung(final Data data) {
		final Data dataVersion = data.getItem("TMCDaten");
		final Data dataVerwaltung = dataVersion.getItem("VerwaltungsInformationen");
		final String status = dataVerwaltung.getItem("TmcStatus").asTextValue().getText();
		return "Gelöscht".equals(status);
	}
}
