/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.situation;

import java.util.List;

import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.tmc.meldungserstellung.AbstraktVewDynObj;

/**
 * Abtrakte Klasse zur Verwaltung von dynamischen DAV-Objekten in Netzen. Diese
 * abstrakte Klasse stellt Funktionalität zur Vefügung um generische Instanzen
 * für dynamische DAV-Objekte, die in dynamischen Mengen von allen Unternetzen
 * eines bestimmeten Typs eines Netzes enthalten sind, zu verwalten<br>
 * . Die verwalteten Instanzen werden in einer (nich modifizierbaren) Map mit
 * der PID des dynamischen Objekts als Schlüssel zur Verfügung gestellt. Diese
 * Map wird entsprechend den Änderungen in der dynamischen Objektmengen und den
 * Löschungen (Invalidierungen) der dynamischen Objekte aktualisiert.<br>
 * Zur Implementierung einer effektiven Verwaltung genügt die Implementierung
 * der Methode <code>erzeugeObjekt</code> und aufrufen der Methode
 * <code>initialisiereNetz</code> z. B. im Konstruktor der beerbenden Klasse.
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 * 
 * @param <T>
 *            Objekt
 */
public abstract class AbstraktNetzVewDynObj<T> extends AbstraktVewDynObj<T> {
	/**
	 * Initialisierte aus dem übergebene Netz - rekursiv - alle enthaltenen
	 * Objekte der angegebenen dynamische Mengen, die durch den übergebenen
	 * Mengentyp bestimmt werden.
	 * 
	 * @param netz
	 *            das Wurzel-Objekt vom Typ typ.netz
	 * @param pidMengenTyp
	 *            der Unterryp des Typs typ.netz, dessen Objekte Besitzer der
	 *            die dynamische Menge sind
	 * @param mengenBezeichner
	 *            Bezeichner (Name) der Menge
	 * @return true, wenn mindestens ein Objekt vom Typ typ.verkehrsModellNetz
	 *         initialisiert wurde, false sonst
	 */
	protected boolean initialisiereNetz(final ConfigurationObject netz, final String pidMengenTyp,
			final String mengenBezeichner) {
		boolean result = false;

		if (!netz.isOfType("typ.netz")) {
			throw new IllegalArgumentException("Objekt " + netz + " ist nicht vom Typ typ.netz");
		}

		if (netz.isOfType(pidMengenTyp)) {
			initialisiere(netz, mengenBezeichner);
			result = true;
		}

		final List<SystemObject> nbt = netz.getNonMutableSet("NetzBestandTeile").getElements();
		if (nbt == null) {
			throw new RuntimeException("Menge \"NetzBestandTeile\" nicht vorhanden in " + netz);
		}

		for (SystemObject obj : nbt) {
			if (obj.isOfType("typ.netz")) {
				result |= initialisiereNetz((ConfigurationObject) obj, pidMengenTyp, mengenBezeichner);
			}
		}
		return result;
	}
}
